;############################################################################
;############################################################################
;##                                                                        ##
;##  AVRTask v0.01 - Simpler Scheduler fr den ATmega162                   ##
;##                                                                        ##
;##	 Christof Krger (c) 2004                                              ##
;##                                                                        ##
;############################################################################
;############################################################################
		.include "m162def.inc"
.cseg
		;Interrupt-Table
		.org 0x0000
		jmp reset		;RESET
		jmp reset		;INT0
		jmp reset		;INT1
		jmp reset		;INT2
		jmp reset		;PCINT0
		jmp reset		;PCINT1
		jmp reset		;TIMER3 CAPT
		jmp reset		;TIMER3 COMPA
		jmp reset		;TIMER3 COMPB
		jmp reset		;TIMER3 OVF
		jmp reset		;TIMER2 COMP
		jmp reset		;TIMER2 OVF
		jmp reset		;TIMER1 CAPT
		jmp reset		;TIMER1 COMPA
		jmp reset		;TIMER1 COMPB
		jmp reset		;TIMER1 OVF
		jmp reset		;TIMER0 COMP
		jmp schedule	;TIMER0 OVF
		jmp reset		;SPI, STC
		jmp reset		;USART0, RXC
		jmp reset		;USART1, RXC
		jmp reset		;USART0, UDRE
		jmp reset		;USART1, URDRE
		jmp reset		;USART0, TXC
		jmp reset		;USART1, TXC
		jmp reset		;EE_RDY
		jmp reset		;ANA_COMP
		jmp reset		;SPM_RDY

		.include "tasks.asm"

;******************************************************************************
reset:
		;Stack initialisieren
		ldi R16, HIGH(RAMEND)
		out SPH, R16
		ldi R16, LOW(RAMEND)
		out SPL, R16

		;Prescaler auf 1/1024 setzen
		; 0 - Aus
		; 1 = Clk
		; 2 - Clk/8
		; 3 - Clk/64
		; 4 - Clk/256
		; 5 - Clk/1024
		ldi R16, 3
		out TCCR0, R16
		;Timer-Interrupt enablen
		ldi R16, (1<<TOIE0)
		out TIMSK, R16

		ldi R16, TASKS-1
		sts current_task, R16

		in R16, MCUCR
		ori R16, (1<<SRE)
		out MCUCR, R16

;******************************************************************************
;Initialisierung der Tasks
;
;Es werden die Init-Prozeduren der einzelnen Tasks nacheinander aufgerufen
;
;******************************************************************************

.dseg
		.org 0x100

current_task:
		.byte 1

z_buffer:		;Puffer, um das Z-Register ohne Hilfe von weiteren Registern mit ldi wiederherstellen zu knnen
		.byte 2  

;taskstatus: ;Noch nicht implementiert
;		.byte 8

taskdata:
		;Offsets
		; 0      - Status-Register
		; 1..32  - R0..R31
		; 33     - PCL
		; 34     - PCH
		; 35     - SPL
		; 36     - SPH
		.equ taskdata_size = 37

		.byte taskdata_size * 8

		.equ taskdata_PCL = 33
		.equ taskdata_PCH = 34
		.equ taskdata_SPL = 35
		.equ taskdata_SPH = 36


.cseg

		.def i = R16			;Zhlvariable
		.def tmp = R17
		.def segmentendL = R24	;Hchste Speicherstelle fr ein Segment, also 
		.def segmentendH = R25	;den Speicherbereich eines Tasks
		.def task_ptr_l = R0	;Adresse des Tasks im Program Memory
		.def task_ptr_h = R1	;Adresse des Tasks im Program Memory
		.def Z_backup_L = R2
		.def Z_backup_H = R3
		.def Main_Adr_L = R4
		.def Main_Adr_H = R5


		;Auslesen der Position der Main-Prozeduren
		ldi i, TASKS
		ldi segmentendH, HIGH(EXTMEM_BASE + TASK_MEM - 1)
		ldi segmentendL, LOW(EXTMEM_BASE + TASK_MEM - 1)
		ldi ZL, LOW(TASKLIST*2)
		ldi ZH, HIGH(TASKLIST*2)
		ldi YL, LOW(taskdata)
		ldi YH, HIGH(taskdata)
init_task:
		lpm task_ptr_l, Z+
		;adiw ZH:ZL, 1
		lpm task_ptr_h, Z+
		movw Z_backup_H:Z_backup_L, ZH:ZL	;Z sichern
		movw ZH:ZL, task_ptr_h:task_ptr_l	;Adresse vom Task in Z laden
		lpm Main_Adr_L, Z+	;Main
		lpm Main_Adr_H, Z+
		std Y+taskdata_PCL, Main_Adr_L		;PC in taskdata schreiben (auf die main-prozedur initialisiert)
		std Y+taskdata_PCH, Main_Adr_H
		std Y+taskdata_SPL, segmentendL
		std Y+taskdata_SPH, segmentendH
					
		movw ZH:ZL, Z_backup_H:Z_backup_L	;Z wiederherstellen

		;SP-Position erhhen
		ldi tmp, LOW(TASK_MEM)
		add segmentendL, tmp
		ldi tmp, HIGH(TASK_MEM)
		adc segmentendH, tmp

		;taskdata-Pointer inkrementieren (Y-Register)
		adiw YL, taskdata_size
		dec i
		brne init_task

;******************************************************************************

		;Ersten Task schedulen
		jmp schedule_out

;******************************************************************************
;******************************************************************************
;******************************************************************************

schedule:
	; #############
	; #  SICHERN  #
	; #############

		;Benutzte Register erstmal auf den Stack sichern
		push R31
		push R30
		push R29
		push R28
		push R0
		push R1
		push R2
		push R16
		push R17
		;Status-Register sichern
		in R2, SREG
		;PC sichern
		in ZL, SPL
		in ZH, SPH
		ldd YH, Z+10
		ldd YL, Z+11
		;Speicheradresse berechnen
		lds R16, current_task
		ldi R17, taskdata_size
		mul R16, R17
		ldi ZL, LOW(taskdata)
		ldi ZH, HIGH(taskdata)
		add ZL, R0
		adc ZH, R1
		;Statusregister schreiben
		st Z+, R2
		;Vorher gebrauchte Register bis auf das Z-Register wieder pop-en
		pop R17
		pop R16
		pop R2
		pop R1
		pop R0
		;Register sichern
		st Z+, R0
		st Z+, R1
		st Z+, R2
		st Z+, R3
		st Z+, R4
		st Z+, R5
		st Z+, R6
		st Z+, R7
		st Z+, R8
		st Z+, R9
		st Z+, R10
		st Z+, R11
		st Z+, R12
		st Z+, R13
		st Z+, R14
		st Z+, R15
		st Z+, R16
		st Z+, R17
		st Z+, R18
		st Z+, R19
		st Z+, R20
		st Z+, R21
		st Z+, R22
		st Z+, R23
		st Z+, R24
		st Z+, R25
		st Z+, R26
		st Z+, R27
		;PC steht noch in Y, also erstmal sichern...
		movw R1:R0, YH:YL
		;...um den vorherigen Inhalt von Y zu speichern
		pop R28
		pop R29
		st Z+, R28
		st Z+, R29
		;Jetzt noch das Z-Register sichern, Adressierung jetzt ber bereits
		;gesichertes X-Register laufen lassen.
		movw XH:XL, ZH:ZL
		pop R30
		pop R31
		st X+, R30
		st X+, R31
		;PC schreiben
		st X+, R0
		st X+, R1
		;SP sichern
		in ZL, SPL
		in ZH, SPH
		adiw ZH:ZL, 2 ;PC berspringen
		st X+, ZL
		st X+, ZH



	; #############
	; #  ABRUFEN  #
	; #############

schedule_out:
		;Nchsten Task whlen (Round-Robin)
		lds R16, current_task
		subi R16, -1
		cpi R16, TASKS
		brlt sched_no_overflow
		;berlauf behandeln
sched_overflow:
		subi R16, TASKS
sched_no_overflow:
		sts current_task, R16

		;Adresse des richtigen taskdata-Abschnittes berechnen (in Z)		
		ldi R17, taskdata_size
		mul R16, R17
		ldi ZL, LOW(taskdata)
		ldi ZH, HIGH(taskdata)
		add ZL, R0
		adc ZH, R1

		;Stackpointer wiederherstellen
		ldd R16, Z+35
		out SPL, R16
		ldd R16, Z+36
		out SPH, R16

		;Rcksprungadresse auf den Stack schreiben
		ldd R0, Z+33
		ldd R1, Z+34
		push R0
		push R1

		;Status-Register wiederherstellen
		ld R16, Z+
		andi R16, 0b01111111
		out SREG, R16

		;Alle Register wiederherstellen
		ldd R0, Z+30
		sts Z_buffer, R0
		ldd R0, Z+31
		sts Z_buffer+1, R0

		ld R0, Z+
		ld R1, Z+
		ld R2, Z+
		ld R3, Z+
		ld R4, Z+
		ld R5, Z+
		ld R6, Z+
		ld R7, Z+
		ld R8, Z+
		ld R9, Z+
		ld R10, Z+
		ld R11, Z+
		ld R12, Z+
		ld R13, Z+
		ld R14, Z+
		ld R15, Z+
		ld R16, Z+
		ld R17, Z+
		ld R18, Z+
		ld R19, Z+
		ld R20, Z+
		ld R21, Z+
		ld R22, Z+
		ld R23, Z+
		ld R24, Z+
		ld R25, Z+
		ld R26, Z+
		ld R27, Z+
		ld R28, Z+
		ld R29, Z+
		lds R30, Z_buffer
		lds R31, Z_buffer+1

		reti

